SOAPAction =
  #
  # Constructor for a SOAPAction object typically
  #  read from a WSDL file.
function(x, obj = new("SOAPAction"))
{
  as(obj, "character") = x

  obj
}


SOAPType =
  #
  # 
  #
function(name, ns = "", nsuri="", namespaceDefs = list(), obj = new("SOAPType"))
{
  if(is.null(name)) 
    return(new("SOAPVoidType"))


   # If the name is in the form of namespace:type, split it
   # and assign to name and ns separately.
   # e.g. xsd:int will become name = "int", ns = "xsd"
  name = strsplit(name, ":")[[1]]
  
  if(length(name) == 2) {
    if(missing(ns))
       ns = name[1]
    name = name[2]        
  }


        # resolve namespace.
  if(missing(nsuri) && length(namespaceDefs)) {
     idx = match(ns, names(namespaceDefs))
     if(is.na(idx) && any(names(namespaceDefs) == ""))
        idx = which(names(namespaceDefs) == "")

     if(!is.na(idx)) {
         nsuri = namespaceDefs[[idx]]$uri
     }
  }

  
  if(nsuri %in% getXSDSchemaURIs()) {
    # tmp = paste(ns, name, sep=":")
    which = sapply(SOAPTypes, function(x) name == x[["type"]])
    if(any(which)) {
       obj = new("PrimitiveSOAPType")
       obj@toConverter = get(paste("as", names(SOAPTypes)[which], sep="."))
       tmp = SOAPTypes[which][[1]]
       if("from" %in% names(tmp)) {

           if(is.character(tmp$from)) {
              if(length(tmp$from) > 1) {
                tmp$from = get(tmp$from[1], paste("package", tmp$from[2], sep = ":"))
              } else
                tmp$from = get(tmp$from)   
           }          
           obj@fromConverter = tmp$from
       }
    }
  } else {
     # Should check that this is referring to the target namespace of the
     # document, e.g. targetNamespace and xmlns:[targetNamespace] in the root node
     obj = new("SOAPTypeReference")
  }

  obj@name = name
  obj@ns = ns
  obj@nsuri = nsuri

  obj
}

SOAPServerDescription =
  # Constructor for a SOAPServerDescription object
  # giving the server identity, and the operations and data types.
function(name,  server, operations, types, obj = new("SOAPServerDescription"))
{
 obj@name = name
 obj@server = server
 if(!is.null(operations))
   obj@operations = operations

 if(!is.null(types))
   obj@types = types

 obj
}


WSDLParseHandlers =
  #
  # Create handlers for reading the WSDL.
  # These handlers discard comments and handle import'ing
  # of referenced schema putting the results back into the
  # tree where the <import> tag was.
  #
function(baseURI, verbose = FALSE, keepSchema = FALSE)
{
  baseURI = dirname(baseURI)
  schema <- list()

  readReferencedSchema =
  function(node) {
          location = xmlGetAttr(node, "schemaLocation", NA)
          if(is.na(location)) {
            warning("skipping ", xmlName(node), " node with no schemaLocation attribute")
            return(NULL)
          }
          
             #XXX Merge the location with the base taking care of absolute addresses, etc.
             # Is this in the XML package?
          location = paste(baseURI, location, sep = "/")
          if(verbose)
             cat(xmlName(node), "ing", location, "as part of", baseURI, "\n")

          doc = xmlRoot(xmlTreeParse(location, handlers = handlers, asTree = TRUE))

          ns = xmlGetAttr(node, "namespace", NA)          
          if(is.na(ns) && xmlName(node) == "import")
              warning("schema <import> without a namespace attribute: ", location)
          else if(keepSchema)
              schema[[ns]] <<- doc
          doc
       }

  
  handlers =
    list(comment = function(node) NULL,
         import = readReferencedSchema,
         include = readReferencedSchema,         
         .schema = function() schema)

  handlers
}  


processWSDL <-
  #
  # read a WSDL file and generate an R object describing the 
  # Web Service that it details.
  #
  # This gets the collection of operations and their parameters and return 
  # types and collects all the type information. It also creates a
  # SOAPServerDescription object to identify the Web server.
  #
  # Currently only handle one <service> element
  #
function(fileName = "KEGG.wsdl", handlers = WSDLParseHandlers(fileName))
{
 library(XML)
    # Parse the WSDL file into a tree which we will pick apart.
    # Could put handlers to convert the elements on the fly.
 doc = xmlRoot(xmlTreeParse(fileName, handlers = handlers,  asTree = TRUE))

    # Get the (first) server declared in <service><port>...
 port = doc[["service"]][["port"]]
 if(sum(xmlSApply(doc[["service"]], xmlName) == "port") > 1)
   warning("Ignoring additional <service><port> ... elements")

   # Get the location/identity of the server.
 loc = xmlGetAttr(port[["address"]], "location")
 server = SOAPServer(loc)

   # Now look at the types section and make up a table of all the types.
 types = processWSDLTypes(doc[["types"]], doc)
 
  # Now find the methods/functions that are available from this server.
  # These are in the <binding> elements. We should match each one to
  # each <service>. Currently only handle one <service> element
 
 tmp = doc[names(doc) == "binding"]
 ops = lapply(tmp, processWSDLBindings, doc, types)
 names(ops) = sapply(tmp, function(x) xmlGetAttr(x, "name"))

 SOAPServerDescription(name = xmlGetAttr(port, "name"),
                       server = server,
                       operations = ops, types = types)
}














processWSDLBindings =
  #
  # Called with a <binding>...</binding> node.
  #
function(node, doc = NULL, typeDefinitions = list())
{
 els =  node[names(node) == "operation"]

    # Get the type attribute and remove the namespace.
 type = discardNamespace(xmlGetAttr(node, "type"))

  # XPath:  portType[@name = type]
 types = doc[xmlSApply(doc, function(x) xmlName(x) == "portType" && xmlGetAttr(x, "name") == type)][[1]]

 namespaceDefinitions = doc$namespaceDefinitions
 # XXX merge with any others for this node.
 
 ops = lapply(els, processWSDLOperation, types, doc, namespaceDefinitions, typeDefinitions)
 names(ops) = sapply(els, xmlGetAttr, "name")
   
 ops  
}

discardNamespace <-
  # utility function that removes the namespace prefix from a string.
  # e.g.  foo:tagName returns tagName.
  #
  # Or  gsub("^.*:", "", type)
function(str)
{
  x = strsplit(str,":")[[1]]
  if(length(x) > 1)
    x[2]
  else
    x
}  

processWSDLOperation =
  #
  # An operation node.
  #
  #
  # types comes from the portType, not the schema.
  # The schema types are in typeDefinitions
  #
  #XXX ignores namespaces.
  #XXX should also get the documentation if there is any (e.g. see nwis).

  #XXX
function(node, types, doc, namespaceDefinitions = list(), typeDefinitions = list())
{
  name = xmlGetAttr(node, "name")

   # Get the operation element within the portType that corresponds to this operation.
  which = xmlSApply(types, function(x) xmlGetAttr(x, "name") == name)

  
  portType = types[which][[1]] # is this equivalent to types[[which]]?
  inputNode = xmlGetAttr(portType[["input"]], "message")
  input = discardNamespace(inputNode)

  
  msgs = doc[xmlSApply(doc, function(x) xmlName(x) == "message" && xmlGetAttr(x, "name") == input)]
  msg = msgs[[1]]

  args = xmlApply(msg, function(x) {
                            el =  xmlGetAttr(x, "type")
                            if(is.null(el)) 
                                el =  xmlGetAttr(x, "element")
                            
                            SOAPType(el, namespaceDefs = namespaceDefinitions)
                          })
   # need to map the name for an element rather than type. Let's do this in
   # genSOAPClientInterface() for now.
  names(args) = xmlSApply(msg, function(x) {
                                     el =  xmlGetAttr(x, "element")
                                     if(is.null(el))
                                       (xmlGetAttr(x, "name"))
                                     else 
                                       resolve(el, typeDefinitions)@name
                               })

  output = xmlGetAttr(portType[["output"]], "message")
  output = discardNamespace(output)
  
  msgs = doc[xmlSApply(doc, function(x) xmlName(x) == "message" && xmlGetAttr(x, "name") == output)]
  msg = msgs[[1]]  


  if(xmlSize(msg)) {
    value = xmlGetAttr(msg[["part"]], "type")
    if(is.null(value)) 
       value = xmlGetAttr(msg[["part"]], "element")

    if(is.null(value))
     value = NULL
  } else
     value = NULL


  namespace = xmlGetAttr(node[["input"]][["body"]], "namespace",
                              xmlGetAttr(doc, "targetNamespace", ""))

  obj = new("WSDLMethod", name = name, parameters = args,
                          returnValue = SOAPType(value, namespaceDefs = namespaceDefinitions),
                          namespace = namespace,
                          action = SOAPAction(xmlGetAttr(node[["operation"]], "soapAction")))

  obj@bindingStyle = xmlGetAttr(node[["operation"]], "style", as.character(NA))
  obj@use = sapply(c("input", "output"), 
                        function(x)
                           xmlGetAttr(node[[x]][["body"]], "use", as.charcter(NA)))

  
   # Deal with case of enclosed within <annotation>
  doc = portType[["documentation"]]
  if(is.null(doc) && !is.null(doc[["annotation"]]))
     doc = doc[["annotation"]][["documentation"]]


  if(!is.null(doc))
    obj@documentation = xmlValue(portType[["documentation"]])

  obj
}  


# Resolve finds the definition and its contents within a context
# It can resolve at one level or all the way down.
# The definitions here are typically SOAP types.

setGeneric("resolve", function(obj, context, namespaces = character(), recursive = TRUE) standardGeneric("resolve"))

setMethod("resolve", c("BasicSOAPType", "list"),
           function(obj, context, namespaces = character(), recursive = TRUE) {
             obj
           })

setMethod("resolve", c("SOAPVoidType"),
           function(obj, context, namespaces = character(), recursive = TRUE) {
             obj
           })

setMethod("resolve", c("SOAPTypeReference", "list"),
           function(obj, context, namespaces = character(), recursive = TRUE) {
             resolve(obj@name, context, namespaces, recursive)
           })

setMethod("resolve", c("ArrayType", "list"),
           function(obj, context, namespaces = character(), recursive = TRUE) {
             if(recursive && (is.null(obj@elType) || length(obj@elType@name) == 0)) {
               name = obj@elementType
               obj@elType = resolve(name, context, namespaces, recursive)
             }
             obj
           })

setMethod("resolve", c("character", "SchemaCollection"),
           function(obj, context, namespaces = character(), recursive = TRUE) {
             for(i in context) {
                ans = try(resolve(obj, i, namespaces, recursive))
                if(!inherits(ans, "try-error"))
                  return(ans)
             }
             stop("Cannot resolve ", obj, " in context")
           })


setMethod("resolve", c("character", "list"),
           function(obj, context, namespaces = character(), recursive = TRUE) {

              if(length(context) == 0)
                stop("Cannot resolve SOAP type in empty context")
              
                  # remove the [] literals and split the string into
                  # namespace prefix & label.
               els = strsplit(gsub("\\[\\]", "", obj), ":")[[1]]
               if(length(els) == 2) {
                 if(els[1] == "xsd") {   #XXX need to match the namespace URI!
                   return(SOAPType(els[2], els[1]))
                 }
                  # Work with just the label, ignoring the namespace prefix.
                 els = els[2]
               } 
             
               el = match(els[1], names(context))
               if(is.na(el)) {
                  el = match(els[1], gsub(".*:", "", names(context)))
               }
               if(is.na(el)) 
                 stop("Cannot resolve SOAPTypeReference ", els[1], " in list")

               ans = context[[el]]$definition

               if(recursive && !is(ans, "TerminalSOAPType"))
                 ans = resolve(ans, context, recursive)

               ans
           })

ArrayType =
function(elementType, namespaceDefs = list(), obj = new("ArrayType"))
{
  return(parseArrayType(elementType, namespaces = namespaceDefs))
}  

ClassDef =
function(name, slotTypes, obj = new("ClassDefinition"))
{
  obj@name = name
  obj@slotTypes = slotTypes

  obj
}

UnionDef =
function(name, slotTypes, obj = new("ClassDefinition"))
{
  ClassDef(name, slotTypes, new("UnionDefinition"))
}

